/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

#ifndef _COMMON_H
#define _COMMON_H

#include "cosm.h"

/* packet types, always the first u32 in the packet */
#define PACKET_TYPE_REQUEST     1
#define PACKET_TYPE_ASSIGNMENT  2
#define PACKET_TYPE_RESULTS     3
#define PACKET_TYPE_ACTIVE      4
#define PACKET_TYPE_ACCEPT      5

/* states for PACKET_ACTIVE */
#define STATE_INVALID         0
#define STATE_ALIVE           1
#define STATE_DONE            2

/* network settings */
#define NETPACKET_SERVER    (ascii *) "localhost" /* full hostname of server */
#define NETPACKET_PORT      (ascii *) "10204"     /* port for communication */

/* client asking for a work assignment */
typedef struct
{
  u32 type;
  u64 speed;
} PACKET_REQUEST;

/* assignment given to client by server */
typedef struct
{
  u32 type;
  /* project assignment data */
  u32 start;
  u32 end;
} PACKET_ASSIGNMENT;

/* results sent to server */
typedef struct
{
  u32 type;
  /* project results */
  u32 start;
  u32 end;
  u64 total;
  u32 cpu;
  u32 os;
  ascii email[32];
} PACKET_RESULTS;

/* ACK/NACK after server accepts results */
typedef struct
{
  u32 type;
  u32 result;
} PACKET_ACCEPT;

/* live work entity, in progress on the client side */
typedef struct
{
  u32 type;
  /* project temporary data */
  u32 start;
  u32 current;
  u32 end;
  u64 total;
  u32 state;
} PACKET_ACTIVE;

#endif
