/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

#ifndef _SERVER_H
#define _SERVER_H

#include "common.h"

/* filenames the server will use */
#define SERVER_DATAFILE (ascii *) "server.dat"
#define SERVER_CONFIG   (ascii *) "server.cfg"
#define WORK_LOG        (ascii *) "work.log"
#define ACTIVITY_LOG    (ascii *) "activity.log"

void CatchSignal( int arg );
  /*
    Sets __shutdown_flag to 1 on signal. In the example it shuts down
    immediately, but with multi-connection or multi-thread versions
    you may want to delay shutdown.
    Returns: nothing.
  */

void Thread( void * arg );
  /*
    A worker thread, handles one connection.
    Returns: nothing.
  */

s32 Initialize( v3_NET * net );
  /*
    This function is responsible for loading in the unassigned work, assigned
    work, and other state information needed for the server. It also sets
    up the network listening port, and any other initial setup.
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

s32 Shutdown( void );
  /*
    This function closes any network connections, and saves all the state data
    such as unassigned work.
    Returns: V3_PASS on success, or V3_FAIL on failure.
  */

void Assign( v3_NET * net );
  /*
    This function takes the open network connection net and reads the
    remainder of a PACKET_REQUEST packet, assigns the client some work, tracks
    the assigned work, and then sends a PACKET_ASSIGNMENT packet to the client
    with the assigned work.
    Returns: nothing.
  */

void Accept( v3_NET * net );
  /*
    This function takes the open network connection net and reads the
    remainder of a PACKET_RESULTS packet, saves the completed work, logs it,
    and then returns a PACKET_ACCEPT packet to the client for acknowledgement.
    Returns: nothing.
  */

int main( int argc, char *argv[] );
  /*
    You know what this is for ;)
  */

#endif
