/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

/* CPU/OS Layer - CPU and OS specific code is allowed */

#include "cosmmath.h"
#include "cosmmem.h"
#include "cosmnet.h"
#include "cosmio.h"

#if ( !defined( NO_NETWORKING ) )
/* #include <string.h> !!! not needed? (audit) */
#if ( OS_TYPE == OS_WIN32 )
#include <windows.h>
#include <stddef.h>
#include <signal.h>
#include <process.h>
#include <time.h>
#include <winsock.h>
#include <io.h>
#define close closesocket
#elif ( OS_TYPE == OS_OS2 )
#include <stddef.h>
#include <signal.h>
#include <process.h>
#include <utils.h>
#include <types.h>
#include <netdb.h>
#include <sys\time.h>
#include <sys\ioctl.h>
#include <sys\socket.h>
#define BSD_SELECT
#include <sys\select.h>
#include <netinet\in.h>
#define close soclose
#else
#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <signal.h>
#endif /* OS */
#endif /* NO_NETWORKING */

/* We need to know if the winsock interface has been initialized */

s32 v3NetOpen( v3_NET * net, u32 host, u16 port, u32 my_host, u16 my_port,
  u32 mode, u32 firehost, u16 fireport, ascii * firepass )
{
#if ( defined( NO_NETWORKING ) )
  if ( net == NULL )
  {
    /* Can't set last_error */
    return( V3_NET_ERROR_NO_NET );
  }
   
  v3MutexLock( &net->lock, V3_MUTEX_WAIT );
  net->last_error = V3_NET_ERROR_NO_NET;
  v3MutexUnlock( &net->lock );
  return( V3_NET_ERROR_NO_NET );
#else
  int socket_descriptor, option;
  /*
    socks4 send_buf: ver + command + port + ip + password (NULL terminated)
     = 1 + 1 + 2 + 4 + [1-64]
  */
  ascii receive_buffer[32], send_buffer[74];
  ascii username[64], password[64];
  u8 send_buffer_length;
  struct sockaddr_in local_address_struct, remote_address_struct;
  int local_address_struct_length, remote_address_struct_length;
  u32 wait_time;
  u32 i, j;
  u16 * temp_u16_ptr;
  u32 * temp_u32_ptr;
  s32 mutex_protected;

  if ( net == NULL )
  {
    /* Can't set last_error */
    return( V3_NET_ERROR_HOST );
  }

  v3MutexLock( &net->lock, V3_MUTEX_WAIT );

  if ( v3_NetGlobalInit() != V3_PASS )
  {
    net->status = V3_NET_STATUS_CLOSED;
    net->last_error = V3_NET_ERROR_NO_NET;
    v3MutexUnlock( &net->lock );
    return( V3_NET_ERROR_NO_NET );
  }

  if ( firehost != 0 )
  {
    /* We cannot (will not) proxy UDP traffic */
    if ( mode == V3_NET_MODE_UDP )
    {
      net->status = V3_NET_STATUS_CLOSED;
      net->last_error = V3_NET_ERROR_MODE;
      v3MutexUnlock( &net->lock );
      return( V3_NET_ERROR_MODE );
    }

    /*
      Using a firewall to connect to host. Temporary way to tell recv
      how long to wait for an answer from the firewall
    */
    wait_time = 120000; /* 120 seconds */

    if ( fireport == 0 )
    {
      /* Invalid port for the firewall */
      net->status = V3_NET_STATUS_CLOSED;
      net->last_error = V3_NET_ERROR_FIREPORT;
      v3MutexUnlock( &net->lock );
      return( V3_NET_ERROR_FIREPORT );
    }
    if ( ( v3u32u64( v3StrLengthA( firepass ) ) + 1 ) >
      sizeof( net->fire_pass ) )
    {
      /* Firewall password is too long */
      net->status = V3_NET_STATUS_CLOSED;
      net->last_error = V3_NET_ERROR_FIREPASS;
      v3MutexUnlock( &net->lock );
      return( V3_NET_ERROR_FIREPASS );
    }

    /* Separating username from password for socks5 firewall */
    for ( i = 0 ; i < v3u32u64( v3StrLengthA( firepass ) ) ; i++ )
    {
      if ( firepass[i] != ';' )
      {
        username[i] = firepass[i];
      }
      else
      {
        username[i] = 0;
        i++;
        break;
      }
    }
    username[i] = 0;

    for ( j = 0 ; i < v3u32u64( v3StrLengthA( firepass ) ) ; i++ )
    {
      if ( firepass[i] == ';' )
      {
        /* Can't have two semi-colon */
        net->status = V3_NET_STATUS_CLOSED;
        net->last_error = V3_NET_ERROR_FIREPASS;
        v3MutexUnlock( &net->lock );
        return( V3_NET_ERROR_FIREPASS );
      }
      password[j] = firepass[i];
      j++;
    }
    password[j] = 0;
    if ( j == 0 )
    {
      /* No semi-colon means socks4 server connection */
      password[j + 1] = 0;
    }
    else
    {
      /* Just in case it's using socks5 with no password */
      password[j + 1] = 1;
    }

    v3MutexUnlock( &net->lock );

    /* Connecting to firewall */
    if ( v3NetOpen( net, firehost, fireport, my_host, my_port,
      V3_NET_MODE_TCP, 0, 0, (ascii *) NULL ) != V3_PASS )
    {
      v3MutexLock( &net->lock, V3_MUTEX_WAIT );
      /* Unable to connect to firewall */
      if ( net->last_error == V3_NET_ERROR_HOST )
      {
        net->last_error = V3_NET_ERROR_FIREHOST;
      }
      mutex_protected = net->last_error;
      v3MutexUnlock( &net->lock );
      return( mutex_protected );
    }

    v3MutexLock( &net->lock, V3_MUTEX_WAIT );

    /* Using socks4 or socks5? */
    if ( ( password[0] != 0 ) || ( password[1] != 0 ) )
    {
      /* Socks5 */
      /* Sending a user/password authentication request for socks5 servers */
      send_buffer[0] = 0x05;
      send_buffer[1] = 0x01;
      send_buffer[2] = 0x02;
      v3MutexUnlock( &net->lock );
      if ( v3NetSend( net, send_buffer, 3 ) != V3_PASS )
      {
        /* Unable to send through the socket */
        v3MutexLock( &net->lock, V3_MUTEX_WAIT );
        net->status = V3_NET_STATUS_CLOSED;
        net->last_error = V3_NET_ERROR_CLOSED;
        v3MutexUnlock( &net->lock );
        return( V3_NET_ERROR_CLOSED );
      }
      if ( v3NetRecv( receive_buffer, 2, net, wait_time ) != 2 )
      {
        /* Didn't receive 2 bytes before timeout or other problem */
        /* Make sure the socket is closed and return */
        v3MutexLock( &net->lock, V3_MUTEX_WAIT );
        if ( net->status != V3_NET_STATUS_CLOSED )
        {
          v3MutexUnlock( &net->lock );
          v3NetClose( net );
          v3MutexLock( &net->lock, V3_MUTEX_WAIT );
        }
        net->last_error = V3_NET_ERROR_CLOSED;
        v3MutexUnlock( &net->lock );
        return( V3_NET_ERROR_CLOSED );
      }
      if ( receive_buffer[1] == 0xFF )
      {
        /* Username/password authentication method refused */
        /* MUST close the connection */
        v3NetClose( net );
        v3MutexLock( &net->lock, V3_MUTEX_WAIT );
        net->status = V3_NET_STATUS_CLOSED;
        net->last_error = V3_NET_ERROR_FIREPASS;
        v3MutexUnlock( &net->lock );
        return( V3_NET_ERROR_FIREPASS );
      }

      v3MutexLock( &net->lock, V3_MUTEX_WAIT );

      if ( receive_buffer[1] == 0x02 )
      {
        /* Username/password authentication method */
        send_buffer[0] = 0x01;
        send_buffer[1] = (u8) v3u32u64( v3StrLengthA( username ) );
        v3MemCopy( &send_buffer[2], username, v3StrLengthA( username ) );
        send_buffer_length =
          (u8) ( v3u32u64( v3StrLengthA( username ) ) + 2 );
        send_buffer[ send_buffer_length ] =
          (u8) v3u32u64( v3StrLengthA( password ) );
        v3MemCopy( &send_buffer[send_buffer_length + 1], password,
          v3StrLengthA( password ) );
        send_buffer_length = (u8) ( send_buffer_length +
        v3u32u64( v3StrLengthA( password ) ) + 1 );

        v3MutexUnlock( &net->lock );

        if ( v3NetSend( net, send_buffer, send_buffer_length ) != V3_PASS )
        {
          /* Unable to send through the socket */
          v3MutexLock( &net->lock, V3_MUTEX_WAIT );
          net->status = V3_NET_STATUS_CLOSED;
          net->last_error = V3_NET_ERROR_CLOSED;
          v3MutexUnlock( &net->lock );
          return( V3_NET_ERROR_CLOSED );
        }
        if ( v3NetRecv( receive_buffer, 2, net, wait_time ) != 2 )
        {
          /* Didn't receive 2 bytes before timeout or other problem */
          /* Make sure the socket is closed and return */
          v3MutexLock( &net->lock, V3_MUTEX_WAIT );
          if ( net->status != V3_NET_STATUS_CLOSED )
          {
            v3MutexUnlock( &net->lock );
            v3NetClose( net );
            v3MutexLock( &net->lock, V3_MUTEX_WAIT );
          }
          net->last_error = V3_NET_ERROR_CLOSED;
          v3MutexUnlock( &net->lock );
          return( V3_NET_ERROR_CLOSED );
        }
        if ( receive_buffer[1] != 0x00 )
        {
          /* Invalid username or password */
          /* server MUST close the connection */
          v3NetClose( net );
          v3MutexLock( &net->lock, V3_MUTEX_WAIT );
          net->last_error = V3_NET_ERROR_FIREPASS;
          v3MutexUnlock( &net->lock );
          return( V3_NET_ERROR_FIREPASS );
        }
        /* Username/password authenticated */
      }
      else if ( receive_buffer[1] != 0x00 )
      {
        /* Error, answer not \0x00 'no authentication necessary' */
        /* nor \0x02 'username/password authentication method'  */
        v3MutexUnlock( &net->lock );
        v3NetClose( net );
        v3MutexLock( &net->lock, V3_MUTEX_WAIT );
        net->last_error = V3_NET_ERROR_FIREPASS;
        v3MutexUnlock( &net->lock );
        return( V3_NET_ERROR_FIREPASS );
      }

      /* It's now time to connect to the remote host */
      send_buffer[0] = 0x05;
      send_buffer[1] = 0x01;
      send_buffer[2] = 0x00;
      send_buffer[3] = 0x01;
      temp_u32_ptr = (u32 *) &send_buffer[4];
      *temp_u32_ptr = htonl( host );
      temp_u16_ptr = (u16 *) &send_buffer[8];
      *temp_u16_ptr = htons( port );
      v3MutexUnlock( &net->lock );
      if ( v3NetSend( net, send_buffer, 10 ) != V3_PASS )
      {
        /* Unable to send through the socket */
        v3MutexLock( &net->lock, V3_MUTEX_WAIT );
        net->status = V3_NET_STATUS_CLOSED;
        net->last_error = V3_NET_ERROR_CLOSED;
        v3MutexUnlock( &net->lock );
        return( V3_NET_ERROR_CLOSED );
      }
      if ( v3NetRecv( receive_buffer, 10, net, wait_time ) != 10 )
      {
        /* Didn't receive 10 bytes before timeout or other problem */
        /* Make sure the socket is closed and return */
        v3MutexLock( &net->lock, V3_MUTEX_WAIT );
        if ( net->status != V3_NET_STATUS_CLOSED )
        {
          v3MutexUnlock( &net->lock );
          v3NetClose( net );
          v3MutexLock( &net->lock, V3_MUTEX_WAIT );
        }
        net->last_error = V3_NET_ERROR_CLOSED;
        v3MutexUnlock( &net->lock );
        return( V3_NET_ERROR_CLOSED );
      }
      if ( receive_buffer[1] != 0x00 )
      {
        /* Connection to host failed */
        /* The server WILL close the connection shortly */
        v3NetClose( net );
        v3MutexLock( &net->lock, V3_MUTEX_WAIT );
        net->last_error = V3_NET_ERROR_HOST;
        v3MutexUnlock( &net->lock );
        return( V3_NET_ERROR_HOST );
      }
      /* The connection has been established */
    }
    else
    {
      /* Using a socks4 server */
      send_buffer[0] = 0x04;
      send_buffer[1] = 0x01;
      temp_u16_ptr = (u16 *) &send_buffer[2];
      *temp_u16_ptr = htons( port );
      temp_u32_ptr = (u32 *) &send_buffer[4];
      *temp_u32_ptr = htonl( host );
      v3StrCopyA( &send_buffer[8], username,
        v3u64u32( v3u32u64( v3StrLengthA( username ) ) + 2 ) );
      v3MutexUnlock( &net->lock );
      if ( v3NetSend( net, send_buffer,
        ( v3u32u64( v3StrLengthA( username ) ) + 9 ) ) != V3_PASS )
      {
        /* Unable to send through the socket */
        v3MutexLock( &net->lock, V3_MUTEX_WAIT );
        net->status = V3_NET_STATUS_CLOSED;
        net->last_error = V3_NET_ERROR_CLOSED;
        v3MutexUnlock( &net->lock );
        return( V3_NET_ERROR_CLOSED );
      }
      if ( v3NetRecv( receive_buffer, 8, net, wait_time ) != 8 )
      {
        /* Didn't receive 8 bytes before timeout or other problem */
        /* Make sure the socket is closed and return */
        v3MutexLock( &net->lock, V3_MUTEX_WAIT );
        if ( net->status != V3_NET_STATUS_CLOSED )
        {
          v3NetClose( net );
          v3MutexUnlock( &net->lock );
        }
        net->last_error = V3_NET_ERROR_CLOSED;
        v3MutexUnlock( &net->lock );
        return( V3_NET_ERROR_CLOSED );
      }
      if ( ( receive_buffer[1] == 92 ) || ( receive_buffer[1] == 93 ) )
      {
        /* Connection rejected */
        v3NetClose( net );
        v3MutexLock( &net->lock, V3_MUTEX_WAIT );
        net->last_error = V3_NET_ERROR_FIREPASS;
        v3MutexUnlock( &net->lock );
        return( V3_NET_ERROR_FIREPASS );
      }
      if ( receive_buffer[1] != 90 )
      {
        /* Connection rejected or failed */
        v3NetClose( net );
        v3MutexLock( &net->lock, V3_MUTEX_WAIT );
        net->last_error = V3_NET_ERROR_HOST;
        v3MutexUnlock( &net->lock );
        return( V3_NET_ERROR_HOST );
      }
      /* The connection has been established */
    }
    v3MutexLock( &net->lock, V3_MUTEX_WAIT );
    net->fire_host = firehost;
    net->fire_port = fireport;
    v3StrCopyA( net->fire_pass, firepass,
      v3u64u32( sizeof( net->fire_pass ) ) );
  }
  else
  {
    /* Direct connection to host */
    if ( mode == V3_NET_MODE_UDP )
    {
      socket_descriptor = socket( AF_INET, SOCK_DGRAM, 0 );
    }
    else
    {
      socket_descriptor = socket( AF_INET, SOCK_STREAM, 0 );
    }
    if ( socket_descriptor < 0 )
    {
      /* unable to open the socket */
      net->status = V3_NET_STATUS_CLOSED;
      net->last_error = V3_NET_ERROR_PORT;
      v3MutexUnlock( &net->lock );
      return( V3_NET_ERROR_PORT );
    }

    local_address_struct_length = sizeof( local_address_struct );
    if ( ( my_host != 0 ) || ( my_port != 0 ) )
    {
      /* Bind to a specified local host/port */
      local_address_struct.sin_family = AF_INET;
      local_address_struct.sin_addr.s_addr = htonl( my_host );
      local_address_struct.sin_port = htons( my_port );

      /* This is to allow us to bind more than one IP to the same port */
      option = 1;
      if ( setsockopt( socket_descriptor, SOL_SOCKET, SO_REUSEADDR,
        (const char *) &option, sizeof( option ) ) < 0 )
      {
        /* unable to set the socket REUSEADDR option */
        net->status = V3_NET_STATUS_CLOSED;
        net->last_error = V3_NET_ERROR_PORT;
        v3MutexUnlock( &net->lock );
        return( V3_NET_ERROR_PORT );
      }

      if ( bind( socket_descriptor, ( struct sockaddr * )
        &local_address_struct, local_address_struct_length ) < 0 )
      {
        /* unable to bind to the socket */
        net->status = V3_NET_STATUS_CLOSED;
        net->last_error = V3_NET_ERROR_PORT;
        v3MutexUnlock( &net->lock );
        return( V3_NET_ERROR_PORT );
      }
    }

    /* Actually connect to remote host */
    remote_address_struct_length = sizeof( remote_address_struct );
    remote_address_struct.sin_family = AF_INET;
    remote_address_struct.sin_addr.s_addr = htonl( host );
    remote_address_struct.sin_port = htons( port );

    if ( connect( socket_descriptor, (struct sockaddr *)
      &remote_address_struct, remote_address_struct_length ) < 0 )
    {
      /* Error connecting to remote host */
      net->status = V3_NET_STATUS_CLOSED;
      net->last_error = V3_NET_ERROR_HOST;
      v3MutexUnlock( &net->lock );
      return( V3_NET_ERROR_HOST );
    }

    /* verify which local address was used to connect to host */
    if ( getsockname( socket_descriptor, ( struct sockaddr * )
      &local_address_struct, &local_address_struct_length ) < 0 )
    {
      /* unable to get the local address and port of the socket */
      net->status = V3_NET_STATUS_CLOSED;
      net->last_error = V3_NET_ERROR_CLOSED;
      v3MutexUnlock( &net->lock );
      return( V3_NET_ERROR_CLOSED );
    }

    net->my_host = ntohl( local_address_struct.sin_addr.s_addr );
    net->my_port = ntohs( local_address_struct.sin_port );
#if ( defined( CPU_64BIT ) )
    net->handle = socket_descriptor;
#else
    net->handle = v3u64u32( socket_descriptor );
#endif
  }

  /* Both direct and through firewall connections set these */
  net->host = host;
  net->port = port;
  net->mode = mode;
  net->status = V3_NET_STATUS_OPEN;
  net->last_error = V3_NET_ERROR_NONE;
  v3MutexUnlock( &net->lock );
  return( V3_PASS );
#endif /* NO_NETWORKING */
}

s32 v3NetSend( v3_NET * net, const void * data, u32 length )
{
#if ( defined( NO_NETWORKING ) )
  if ( net == NULL )
  {
    return( V3_NET_ERROR_NO_NET );
  }

  v3MutexLock( &net->lock, V3_MUTEX_WAIT );
  net->last_error = V3_NET_ERROR_NO_NET;
  v3MutexUnlock( &net->lock );
  return( V3_NET_ERROR_NO_NET );
#else
  int socket_descriptor;
  int flags;
  int result;
  u32 sent, to_send;
  u32 packet_size;
  u8 * buffer;

  if ( net == NULL )
  {
    return( V3_NET_ERROR_CLOSED );
  }

  if ( ( data == NULL ) || ( length == 0 ) )
  {
    /* no data to send is OK */
    return( V3_PASS );
  }

  v3MutexLock( &net->lock, V3_MUTEX_WAIT );

  if ( net->mode != V3_NET_MODE_TCP )
  {
    net->last_error = V3_NET_ERROR_MODE;
    v3MutexUnlock( &net->lock );
    return( 0 );
  }

  if ( v3_NetGlobalInit() != V3_PASS )
  {
    net->status = V3_NET_STATUS_CLOSED;
    net->last_error = V3_NET_ERROR_NO_NET;
    v3MutexUnlock( &net->lock );
    return( V3_NET_ERROR_NO_NET );
  }

#if ( defined( CPU_64BIT ) )
  socket_descriptor = net->handle;
#else
  socket_descriptor = v3u32u64( net->handle );
#endif

  flags = 0;
  sent = 0;
  buffer = (u8 *) data;

  /*
    this is to make sure we dont go and send packets that are difficult
    for the routers to handle (not to mention firewalls)
  */
  packet_size = 49152; /* 48kB */

  /* we will now send the data in chunks of packet_size bytes */
  while ( sent < length )
  {
    if ( ( length - sent ) < packet_size )
    {
      to_send = length - sent;
    }
    else
    {
      to_send = packet_size;
    }

    result = send( socket_descriptor, (const char *) buffer, to_send, flags );
 
    if ( result == 0 )
    {
      /* Connection has been closed */
      net->status = V3_NET_STATUS_CLOSED;
      net->last_error = V3_NET_ERROR_CLOSED;
      v3MutexUnlock( &net->lock );
      return( V3_NET_ERROR_CLOSED );
    }
    if ( result < 0 )
    {
      /* !!! this needs work */
      /* We don't have many types of errors available */
      net->last_error = V3_NET_ERROR_CLOSED;
      v3MutexUnlock( &net->lock );
      return( V3_NET_ERROR_CLOSED );
    }
    sent = sent + result;
    buffer = &buffer[result];
  }
  /* Complete buffer sent */
  v3MutexUnlock( &net->lock );
  return( V3_PASS );
#endif /* NO_NETWORKING */
}

u32 v3NetRecv( void * buffer, u32 length, v3_NET * net, u32 wait_ms )
{
#if ( defined( NO_NETWORKING ) )
  return( 0 );
#else
  int socket_descriptor, flags, result;
  u32 received;
  struct timeval time_struct;
  fd_set descriptors_settings;
  u8 * data;
  v3_time time_called, time_now, time_elapsed;
  u64 tmp_u64;

  if ( net == NULL )
  {
    /* Can't set last_error */
    return( 0 );
  }

  v3MutexLock( &net->lock, V3_MUTEX_WAIT );

  if ( net->mode != V3_NET_MODE_TCP )
  {
    net->last_error = V3_NET_ERROR_MODE;
    v3MutexUnlock( &net->lock );
    return( 0 );
  }

  if ( v3_NetGlobalInit() != V3_PASS )
  {
    net->status = V3_NET_STATUS_CLOSED;
    net->last_error = V3_NET_ERROR_NO_NET;
    v3MutexUnlock( &net->lock );
    return( 0 );
  }

  if ( buffer == NULL )
  {
    net->last_error = V3_NET_ERROR_HOST;
    v3MutexUnlock( &net->lock );
    return( 0 );
  }

#if ( defined( CPU_64BIT ) )
  socket_descriptor = net->handle;
#else
  socket_descriptor = v3u32u64( net->handle );
#endif

  /* Set the socket as non-blocking */
#if ( OS_TYPE == OS_OS2 )
  flags = -1; /* set nonblocking mode */
  if ( ioctl( socket_descriptor, FIONBIO, (char *) &flags,
    sizeof( flags ) ) < 0 )
  {
    /* Unable to set flags of the socket to non-blocking */
    net->last_error = V3_NET_ERROR_HOST;
    v3MutexUnlock( &net->lock );
    return( 0 );
  }
#endif

  data = (u8 *) buffer;
  received = 0;
  flags = 0;
  FD_ZERO( &descriptors_settings );
  FD_SET( socket_descriptor, &descriptors_settings );

  /*
   The timeval struct gets modified by select under Linux to represent
   the remaining time when it returns but remains unchanged on other
   Unices / POSIX.1g so we'll have to use one of the time functions
   to modify the struct ourselves on the second and following calls
  */
  if ( v3SystemClock( &time_called ) != V3_PASS )
  {
    /* Can't get local time */
    net->last_error = V3_NET_ERROR_HOST;
    v3MutexUnlock( &net->lock );
    return( 0 );
  }

  while ( received < length )
  {
    if ( v3SystemClock( &time_now ) != V3_PASS )
    {
      /* Can't get local time */
      net->last_error = V3_NET_ERROR_HOST;
      v3MutexUnlock( &net->lock );
      return( received );
    }

    /* Lets fill that time_struct for the remaining time to wait */
    time_elapsed = v3s128Sub( time_now, time_called );
    _V3_SET64( tmp_u64, 000010C6, F7A0B5ED );
    time_struct.tv_sec = ( wait_ms / 1000 ) - v3s32s64( time_elapsed.hi );
    time_struct.tv_usec = ( ( wait_ms % 1000 ) * 1000 ) -
      v3u32u64( v3u64Div( time_elapsed.lo, tmp_u64 ) );

    if ( time_struct.tv_usec < 0 )
    {
      time_struct.tv_usec = time_struct.tv_usec + 1000000;
      time_struct.tv_sec--;
    }
    if ( time_struct.tv_sec < 0 )
    {
      /* Time is now elapsed, just take what's in the buffer and exit */
      time_struct.tv_sec = 0;
      time_struct.tv_usec = 0;
    }

    result = select( socket_descriptor + 1, &descriptors_settings,
      (fd_set *) NULL, (fd_set *) NULL, &time_struct );
    if ( result < 1 )
    {
      if ( result < 0 )
      {
        /* Error with select */
        net->last_error = V3_NET_ERROR_HOST;
      }
      else
      {
        /* Timeout */
        net->last_error = V3_NET_ERROR_WAIT;
      }
      v3MutexUnlock( &net->lock );
      return( received );
    }
    result = recv( socket_descriptor, (char *) &data[received],
      length - received, flags );
    if ( result < 1 )
    {
      if ( result == 0 )
      {
        /* Connection terminated */
        net->status = V3_NET_STATUS_CLOSED;
        net->last_error = V3_NET_ERROR_CLOSED;
      }
      else
      {
        /* Error while receiving data */
        net->last_error = V3_NET_ERROR_HOST;
      }
      v3MutexUnlock( &net->lock );
      return( received );
    }
    received = received + result;
  }
  /* Received all length bytes */
  v3MutexUnlock( &net->lock );
  return( received );
#endif /* NO_NETWORKING */
}

s32 v3NetSendUDP( v3_NET * net, const void * data, u32 length )
{
#if ( defined( NO_NETWORKING ) )
  if ( net == NULL )
  {
    return( V3_NET_ERROR_NO_NET );
  }
  
  v3MutexLock( &net->lock, V3_MUTEX_WAIT );
  net->last_error = V3_NET_ERROR_NO_NET;
  v3MutexUnlock( &net->lock );
  return( V3_NET_ERROR_NO_NET );
#else
  int socket_descriptor;
  int flags;
  int result;
  struct sockaddr_in to_addr;

  if ( net == NULL )
  {
    return( V3_NET_ERROR_CLOSED );
  }

  if ( ( data == NULL ) || ( length == 0 ) )
  {
    /* no data to send is OK */
    return( V3_PASS );
  }

  v3MutexLock( &net->lock, V3_MUTEX_WAIT );

  if ( net->mode != V3_NET_MODE_UDP )
  {
    net->last_error = V3_NET_ERROR_MODE;
    v3MutexUnlock( &net->lock );
    return( 0 );
  }

  if ( v3_NetGlobalInit() != V3_PASS )
  {
    net->status = V3_NET_STATUS_CLOSED;
    net->last_error = V3_NET_ERROR_NO_NET;
    v3MutexUnlock( &net->lock );
    return( V3_NET_ERROR_NO_NET );
  }

#if ( defined( CPU_64BIT ) )
  socket_descriptor = net->handle;
#else
  socket_descriptor = v3u32u64( net->handle );
#endif

  flags = 0;

  /* Port / Address of remote host */
  to_addr.sin_family = AF_INET;
  to_addr.sin_addr.s_addr = htonl( net->host );
  to_addr.sin_port = htons( net->port );

  /* Sends using UDP are atomic (in the network sense ). So this */
  /* logic is much more simple compared to its TCP counterpart.  */

  result = sendto( socket_descriptor, (const char *) data, length, flags,
                   (struct sockaddr *) &to_addr, sizeof( to_addr ) );

  if ( result == 0 )
  {
    /* Connection has been closed */
    net->status = V3_NET_STATUS_CLOSED;
    net->last_error = V3_NET_ERROR_CLOSED;
    v3MutexUnlock( &net->lock );
    return( V3_NET_ERROR_CLOSED );
  }
  if ( result < 0 )
  {
    /* !!! this needs work */
    /* We don't have many types of errors available */
    net->last_error = V3_NET_ERROR_HOST;
    v3MutexUnlock( &net->lock );
    return( V3_NET_ERROR_HOST );
  }

  /* Complete buffer sent */
  v3MutexUnlock( &net->lock );
  return( V3_PASS );
#endif
}

u32 v3NetRecvUDP( void * buffer, u32 length, v3_NET * net, v3_NET_ACL * acl,
  u32 wait_ms )
{
#if ( defined( NO_NETWORKING ) )
  if ( net == NULL )
  {
    return( 0 );
  }

  v3MutexLock( &net->lock, V3_MUTEX_WAIT );
  net->last_error = V3_NET_ERROR_NO_NET;
  v3MutexUnlock( &net->lock );
  return( 0 );
#else
  int socket_descriptor, flags, result;
  u32 received;
  struct timeval time_struct;
  fd_set descriptors_settings;
  u8 * data;
  v3_time time_called, time_now, time_elapsed;
  u64 tmp_u64;
  s32 acl_test_result;
  struct sockaddr_in client_address_struct;
  int client_address_struct_length;

  if ( net == NULL )
  {
    return( 0 );
  }

  v3MutexLock( &net->lock, V3_MUTEX_WAIT );

  if ( net->mode != V3_NET_MODE_UDP )
  {
    net->last_error = V3_NET_ERROR_MODE;
    v3MutexUnlock( &net->lock );
    return( 0 );
  }

  if ( v3_NetGlobalInit() != V3_PASS )
  {
    net->status = V3_NET_STATUS_CLOSED;
    net->last_error = V3_NET_ERROR_NO_NET;
    v3MutexUnlock( &net->lock );
    return( (u32) V3_NET_ERROR_NO_NET );
  }

  if ( buffer == NULL )
  {
    net->last_error = V3_NET_ERROR_HOST;
    v3MutexUnlock( &net->lock );
    return( 0 );
  }

#if ( defined( CPU_64BIT ) )
  socket_descriptor = net->handle;
#else
  socket_descriptor = v3u32u64( net->handle );
#endif

  /* Set the socket as non-blocking */
#if ( OS_TYPE == OS_OS2 )
  flags = -1; /* set nonblocking mode */
  if ( ioctl( socket_descriptor, FIONBIO, (char *) &flags,
    sizeof( flags ) ) < 0 )
  {
    /* Unable to set flags of the socket to non-blocking */
    net->last_error = V3_NET_ERROR_HOST;
    v3MutexUnlock( &net->lock );
    return( 0 );
  }
#endif

  data = (u8 *) buffer;
  received = 0;
  flags = 0;
  FD_ZERO( &descriptors_settings );
  FD_SET( socket_descriptor, &descriptors_settings );

  /*
   The timeval struct gets modified by select under Linux to represent
   the remaining time when it returns but remains unchanged on other
   Unices / POSIX.1g so we'll have to use one of the time functions
   to modify the struct ourselves on the second and following calls
  */
  if ( v3SystemClock( &time_called ) != V3_PASS )
  {
    /* Can't get local time */
    net->last_error = V3_NET_ERROR_HOST;
    v3MutexUnlock( &net->lock );
    return( 0 );
  }

  while ( received < length )
  {
    if ( v3SystemClock( &time_now ) != V3_PASS )
    {
      /* Can't get local time */
      net->last_error = V3_NET_ERROR_HOST;
      v3MutexUnlock( &net->lock );
      return( received );
    }

    /* Lets fill that time_struct for the remaining time to wait */
    time_elapsed = v3s128Sub( time_now, time_called );
    _V3_SET64( tmp_u64, 000010C6, F7A0B5ED );
    time_struct.tv_sec = ( wait_ms / 1000 ) - v3s32s64( time_elapsed.hi );
    time_struct.tv_usec = ( ( wait_ms % 1000 ) * 1000 ) -
      v3u32u64( v3u64Div( time_elapsed.lo, tmp_u64 ) );

    if ( time_struct.tv_usec < 0 )
    {
      time_struct.tv_usec = time_struct.tv_usec + 1000000;
      time_struct.tv_sec--;
    }
    if ( time_struct.tv_sec < 0 )
    {
      /* Time is now elapsed, just take what's in the buffer and exit */
      time_struct.tv_sec = 0;
      time_struct.tv_usec = 0;
    }

    result = select( socket_descriptor + 1, &descriptors_settings,
      (fd_set *) NULL, (fd_set *) NULL, &time_struct );
    if ( result < 1 )
    {
      if ( result < 0 )
      {
        /* Error with select */
        net->last_error = V3_NET_ERROR_HOST;
      }
      else
      {
        /* Timeout */
        net->last_error = V3_NET_ERROR_WAIT;
      }
      v3MutexUnlock( &net->lock );
      return( received );
    }

    client_address_struct_length = sizeof( client_address_struct );

    result = recvfrom( socket_descriptor, (char *) &data[received],
      length - received, flags, (struct sockaddr *) &client_address_struct,
      &client_address_struct_length );

    if ( result < 1 )
    {
      if ( result == 0 )
      {
        /* Connection terminated */
        net->status = V3_NET_STATUS_CLOSED;
        net->last_error = V3_NET_ERROR_CLOSED;
      }
      else
      {
        /* Error while receiving data */
        net->last_error = V3_NET_ERROR_HOST;
      }
      v3MutexUnlock( &net->lock );
      return( received );
    }

    acl_test_result = v3NetACLTest( acl,
                           ntohl( client_address_struct.sin_addr.s_addr ) );

    /* only accept this data if host passes acl masks */
    if (acl_test_result == V3_NET_ALLOW)
    {
      received = received + result;
    }
  }
  /* Received all length bytes */
  v3MutexUnlock( &net->lock );
  return( received );
#endif
}

s32 v3NetListen( v3_NET * net, u32 ip, u16 port, u32 mode, u32 queue )
{
#if ( defined( NO_NETWORKING ) )
  if ( net == NULL )
  {
    /* Can't set last_error */
    return( V3_NET_ERROR_NO_NET );
  }
  
  v3MutexLock( &net->lock, V3_MUTEX_WAIT );
  net->last_error = V3_NET_ERROR_NO_NET;
  v3MutexUnlock( &net->lock );
  return( V3_NET_ERROR_NO_NET );
#else
  int socket_descriptor, option;
  struct sockaddr_in socket_address_struct;
  int socket_address_struct_length;

  if ( net == NULL )
  {
    /* Can't set last_error */
    return( V3_NET_ERROR_HOST );
  }

  v3MutexLock( &net->lock, V3_MUTEX_WAIT );

  if ( v3_NetGlobalInit() != V3_PASS )
  {
    net->status = V3_NET_STATUS_CLOSED;
    net->last_error = V3_NET_ERROR_NO_NET;
    v3MutexUnlock( &net->lock );
    return( V3_NET_ERROR_NO_NET );
  }

  if ( mode == V3_NET_MODE_UDP )
  {
    socket_descriptor = socket( AF_INET, SOCK_DGRAM, 0 );
  }
  else
  {
    socket_descriptor = socket( AF_INET, SOCK_STREAM, 0 );
  }
  if ( socket_descriptor < 0 )
  {
    /* unable to open the socket */
    net->status = V3_NET_STATUS_CLOSED;
    net->last_error = V3_NET_ERROR_PORT;
    v3MutexUnlock( &net->lock );
    return( V3_NET_ERROR_PORT );
  }

  socket_address_struct.sin_family = AF_INET;
  socket_address_struct.sin_addr.s_addr = htonl( ip );
  socket_address_struct.sin_port = htons( port );
  socket_address_struct_length = sizeof( socket_address_struct );

  option = 1;
#if ( OS_TYPE == OS_FREEBSD )
  /*
    This is to allow us to bind more than one IP to the same port
    Linux and Solaris don't need it and at least one machine running
    linux 2.2.10-ac6 had problem with it so it's only used on those
    platforms that needs it (if necessary, the code will be in
    the #else part)

    Actually, this is necessary to restart a service with connections
    still in the process of shutdown - cramer
  */
  if ( setsockopt( socket_descriptor, SOL_SOCKET, SO_REUSEADDR,
    ( void * ) &option, sizeof( option ) ) < 0 )
  {
    /* unable to set the socket REUSEADDR option */
    net->status = V3_NET_STATUS_CLOSED;
    net->last_error = V3_NET_ERROR_PORT;
    v3MutexUnlock( &net->lock );
    return( V3_NET_ERROR_PORT );
  }
#endif

  if ( bind( socket_descriptor, ( struct sockaddr * ) &socket_address_struct,
    socket_address_struct_length ) < 0 )
  {
    /* unable to bind to the socket */
    net->status = V3_NET_STATUS_CLOSED;
    net->last_error = V3_NET_ERROR_PORT;
    v3MutexUnlock( &net->lock );
    return( V3_NET_ERROR_PORT );
  }

  if ( listen( socket_descriptor, queue ) < 0 )
  {
    /* unable to listen to the socket */
    net->status = V3_NET_STATUS_CLOSED;
    net->last_error = V3_NET_ERROR_PORT;
    v3MutexUnlock( &net->lock );
    return( V3_NET_ERROR_PORT );
  }

  if ( port == 0 )
  {
    if ( getsockname( socket_descriptor, ( struct sockaddr * )
      &socket_address_struct, &socket_address_struct_length ) < 0 )
    {
      /* unable to get the port the socket is listenning on */
      net->status = V3_NET_STATUS_CLOSED;
      net->last_error = V3_NET_ERROR_PORT;
      v3MutexUnlock( &net->lock );
      return( V3_NET_ERROR_PORT );
    }
    net->my_port = ntohs( socket_address_struct.sin_port );
  }
  else
  {
    net->my_port = port;
  }

  net->my_host = ip;

#if ( defined( CPU_64BIT ) )
  net->handle = socket_descriptor;
#else
  net->handle = v3u64u32( socket_descriptor );
#endif

  net->mode = mode;
  net->status = V3_NET_STATUS_LISTEN;
  net->last_error = V3_NET_ERROR_NONE;
  v3MutexUnlock( &net->lock );
  return( V3_PASS );
#endif /* NO_NETWORKING */
}

s32 v3NetAccept( v3_NET * new_connection, v3_NET * net, v3_NET_ACL * acl,
  u32 wait )
{
#if ( defined( NO_NETWORKING ) )
  if ( new_connection == NULL )
  {
    return( V3_NET_ERROR_NO_NET );
  }

  if ( net == NULL )
  {
    v3MutexLock( &new_connection->lock, V3_MUTEX_WAIT );
    new_connection->last_error = V3_NET_ERROR_NO_NET;
    v3MutexUnlock( &new_connection->lock );
    return( V3_NET_ERROR_NO_NET );
  }

  v3MutexLock( &net->lock, V3_MUTEX_WAIT );
  net->last_error = V3_NET_ERROR_NO_NET;
  v3MutexUnlock( &net->lock );
  return( V3_NET_ERROR_NO_NET );
#else
  int socket_descriptor, new_socket_descriptor;
  struct sockaddr_in client_address_struct, local_address_struct;
  int client_address_struct_length, local_address_struct_length;
  fd_set descriptors_settings;
  struct timeval time_struct;
  int flags, result;
  s32 acl_test_result;

  if ( new_connection == NULL )
  {
    /* Can't set last_error */
    return( V3_NET_ERROR_HOST );
  }

  if ( net == NULL )
  {
    v3MutexLock( &new_connection->lock, V3_MUTEX_WAIT );
    new_connection->last_error = V3_NET_ERROR_HOST;
    v3MutexUnlock( &new_connection->lock );
    return( V3_NET_ERROR_HOST );
  }

  v3MutexLock( &net->lock, V3_MUTEX_WAIT );

  if ( v3_NetGlobalInit() != V3_PASS )
  {
    net->status = V3_NET_STATUS_CLOSED;
    net->last_error = V3_NET_ERROR_NO_NET;
    v3MutexUnlock( &net->lock );
    return( V3_NET_ERROR_NO_NET );
  }

  v3MutexLock( &new_connection->lock, V3_MUTEX_WAIT );

  /* v3NetAccept is not valid for UDP connections */
  if ( net->mode == V3_NET_MODE_UDP )
  {
    net->last_error = V3_NET_ERROR_MODE;
    new_connection->status = V3_NET_STATUS_CLOSED;
    new_connection->last_error = V3_NET_ERROR_MODE;
    v3MutexUnlock( &new_connection->lock );
    v3MutexUnlock( &net->lock );
    return( V3_NET_ERROR_MODE );
  }

  flags = 0;

#if ( defined( CPU_64BIT ) )
  socket_descriptor = net->handle;
#else
  socket_descriptor = v3u32u64( net->handle );
#endif

  if ( wait == V3_NET_ACCEPT_NOWAIT )
  {
    /* We want select to return immediately */
    time_struct.tv_sec = 0;
    time_struct.tv_usec = 0;

    FD_ZERO( &descriptors_settings );
    FD_SET( socket_descriptor, &descriptors_settings );
    result = select( socket_descriptor + 1, &descriptors_settings,
      (fd_set *) NULL, (fd_set *) NULL, &time_struct );

    if ( result < 1 )
    {
      if ( result < 0 )
      {
        /* Error with select */
        new_connection->last_error = V3_NET_ERROR_CLOSED;
      }
      else
      {
        /* No connection waiting */
        new_connection->last_error = V3_NET_ERROR_CLOSED;
      }
      new_connection->status = V3_NET_STATUS_CLOSED;
      v3MutexUnlock( &new_connection->lock );
      v3MutexUnlock( &net->lock );
      return( new_connection->last_error );
    }

    /* Making sure we won't block on accept() if the connection is lost */
#if ( OS_TYPE == OS_WIN32 )
    /* !!! */
#elif ( OS_TYPE == OS_OS2 )
    flags = -1; /* set nonblocking mode */
    if ( ioctl( socket_descriptor, FIONBIO, (char *) &flags,
      sizeof( flags ) ) < 0 )
    {
      /* Unable to set flags of the socket to non-blocking */
      new_connection->status = V3_NET_STATUS_CLOSED;
      new_connection->last_error = V3_NET_ERROR_CLOSED;
      v3MutexUnlock( &new_connection->lock );
      v3MutexUnlock( &net->lock );
      return( V3_NET_ERROR_CLOSED );
    }
#else
    /* Get the socket's flags */
    flags = fcntl( socket_descriptor, F_GETFL, 0 );
    if ( flags < 0 )
    {
      /* Unable to get flags of the socket */
      net->status = V3_NET_STATUS_CLOSED;
      net->last_error = V3_NET_ERROR_CLOSED;
      new_connection->status = V3_NET_STATUS_CLOSED;
      new_connection->last_error = V3_NET_ERROR_CLOSED;
      v3MutexUnlock( &new_connection->lock );
      v3MutexUnlock( &net->lock );
      return( V3_NET_ERROR_CLOSED );
    }

    /* Change the socket's flags - Turn off blocking */
    flags = flags | O_NONBLOCK;
    if ( fcntl( socket_descriptor, F_SETFL, flags ) < 0 )
    {
      /* Unable to set flags of the socket to non-blocking */
      net->status = V3_NET_STATUS_CLOSED;
      net->last_error = V3_NET_ERROR_CLOSED;
      new_connection->status = V3_NET_STATUS_CLOSED;
      new_connection->last_error = V3_NET_ERROR_CLOSED;
      v3MutexUnlock( &new_connection->lock );
      v3MutexUnlock( &net->lock );
      return( V3_NET_ERROR_CLOSED );
    }
#endif
  }

  client_address_struct_length = sizeof( client_address_struct );

  /* we loop if wait != V3_NET_ACCEPT_NOWAIT and ACL failed. */
  do
  {
    /* accept() the connection ... */
    new_socket_descriptor = accept( socket_descriptor, (struct sockaddr *)
      &client_address_struct, &client_address_struct_length );

    /* if we connected ... */
    if ( new_socket_descriptor > 0 )
    {
      /* Make sure the host is allowed by ACL masks */
      acl_test_result = v3NetACLTest( acl,
                          ntohl( client_address_struct.sin_addr.s_addr ) );

      if ( acl_test_result == V3_NET_DENY )
      {
        close( new_socket_descriptor );
      }
    }
    else
    {
      acl_test_result = V3_NET_DENY;
    }
  } while ( acl_test_result == V3_NET_DENY && wait != V3_NET_ACCEPT_NOWAIT );

  if ( wait == V3_NET_ACCEPT_NOWAIT )
  {
    /*
      The old socket need to be set blocking again in case the next
      call to v3NetListen() is with V3_NET_ACCEPT_WAIT
    */
#if ( OS_TYPE == OS_WIN32 )
    /* !!! */
#elif ( OS_TYPE == OS_OS2 )
    flags = 0; /* (re)set blocking mode */
    if ( ioctl( socket_descriptor, FIONBIO, (char *) &flags,
      sizeof( flags ) ) < 0 )
    {
      /* Unable to reset flags of the listenning socket to blocking */
      new_connection->status = V3_NET_STATUS_CLOSED;
      new_connection->last_error = V3_NET_ERROR_CLOSED;
      v3MutexUnlock( &new_connection->lock );
      v3MutexUnlock( &net->lock );
      return( V3_NET_ERROR_CLOSED );
    }
    if ( new_socket_descriptor > 0 )
    {
      if ( ioctl( new_socket_descriptor, FIONBIO, (char *) &flags,
        sizeof( flags ) ) < 0 )
      {
        /* Unable to set flags of the new socket to blocking */
        new_connection->status = V3_NET_STATUS_CLOSED;
        new_connection->last_error = V3_NET_ERROR_CLOSED;
        v3MutexUnlock( &new_connection->lock );
        v3MutexUnlock( &net->lock );
        return( V3_NET_ERROR_CLOSED );
      }
    }
#else
    flags = flags & ~O_NONBLOCK;
    if ( fcntl( socket_descriptor, F_SETFL, flags ) < 0 )
    {
      /* Unable to reset flags of the listenning socket to blocking */
      net->status = V3_NET_STATUS_CLOSED;
      net->last_error = V3_NET_ERROR_CLOSED;
      /* The new connection may still be open, lets continue */
    }
#endif
  }

  /* We have nothing more to do with net */
  v3MutexUnlock( &net->lock );

  /* check that we got a connection */
  if ( new_socket_descriptor < 0 )
  {
    /* Unable to accept a client - some error ? */
    /* On NOWAIT, may be a lost connection between select() and accept() */
    new_connection->status = V3_NET_STATUS_CLOSED;
    new_connection->last_error = V3_NET_ERROR_CLOSED;
    v3MutexUnlock( &new_connection->lock );
    return( V3_NET_ERROR_CLOSED );
  }

  if ( wait == V3_NET_ACCEPT_NOWAIT )
  {
    /* The new socket must also be set to blocking */
#if ( OS_TYPE == OS_WIN32 )
    /* !!!  */
#elif ( OS_TYPE == OS_OS2 )
    /* Change the new socket's flags - Turn on blocking */
    flags = 0;
    if ( ioctl( new_socket_descriptor, FIONBIO, (char *) &flags,
      sizeof( flags ) ) < 0 )
    {
      /* Unable to set flags of the new socket to blocking */
      new_connection->status = V3_NET_STATUS_CLOSED;
      new_connection->last_error = V3_NET_ERROR_CLOSED;
      v3MutexUnlock( &new_connection->lock );
      return( V3_NET_ERROR_CLOSED );
    }
#else
    /* Get the new socket's flags */
    flags = fcntl( new_socket_descriptor, F_GETFL, 0 );
    if ( flags < 0 )
    {
      /* Unable to get flags of the new socket */
      new_connection->status = V3_NET_STATUS_CLOSED;
      new_connection->last_error = V3_NET_ERROR_CLOSED;
      v3MutexUnlock( &new_connection->lock );
      return( V3_NET_ERROR_CLOSED );
    }

    /* Change the new socket's flags - Turn on blocking */
    flags = flags & ~O_NONBLOCK;
    if ( fcntl( new_socket_descriptor, F_SETFL, flags ) < 0 )
    {
      /* Unable to set flags of the new socket to blocking */
      new_connection->status = V3_NET_STATUS_CLOSED;
      new_connection->last_error = V3_NET_ERROR_CLOSED;
      v3MutexUnlock( &new_connection->lock );
      return( V3_NET_ERROR_CLOSED );
    }
#endif
  }

  /* verify which address the client has connected to */
  local_address_struct_length = sizeof( local_address_struct );
  if ( getsockname( new_socket_descriptor, ( struct sockaddr * )
    &local_address_struct, &local_address_struct_length ) < 0 )
  {
    /* unable to get the local address and port of the socket */
    new_connection->status = V3_NET_STATUS_CLOSED;
    new_connection->last_error = V3_NET_ERROR_CLOSED;
    v3MutexUnlock( &new_connection->lock );
    return( V3_NET_ERROR_CLOSED );
  }

  /* Filling the new connection v3_NET struct */
  new_connection->my_host = ntohl( local_address_struct.sin_addr.s_addr );
  new_connection->my_port = ntohs( local_address_struct.sin_port );
  new_connection->host = ntohl( client_address_struct.sin_addr.s_addr );
  new_connection->port = ntohs( client_address_struct.sin_port );
  new_connection->mode = V3_NET_MODE_TCP;

#if ( defined( CPU_64BIT ) )
  new_connection->handle = new_socket_descriptor;
#else
  new_connection->handle = v3u64u32( new_socket_descriptor );
#endif

  new_connection->status = V3_NET_STATUS_OPEN;
  new_connection->last_error = V3_NET_ERROR_NONE;
  v3MutexUnlock( &new_connection->lock );
  return( V3_PASS );
#endif /* NO_NETWORKING */
}

s32 v3NetClose( v3_NET * net )
{
#if ( defined( NO_NETWORKING ) )
  if ( net == NULL )
  {
    return( V3_NET_ERROR_NO_NET );
  }
   
  v3MutexLock( &net->lock, V3_MUTEX_WAIT );
  net->last_error = V3_NET_ERROR_NO_NET;
  v3MutexUnlock( &net->lock );
  return( V3_NET_ERROR_NO_NET );
#else
  int socket_descriptor;

  if ( net == NULL )
  {
    return( V3_NET_ERROR_CLOSED );
  }

  v3MutexLock( &net->lock, V3_MUTEX_WAIT );

  if ( v3_NetGlobalInit() != V3_PASS )
  {
    net->status = V3_NET_STATUS_CLOSED;
    net->last_error = V3_NET_ERROR_NO_NET;
    v3MutexUnlock( &net->lock );
    return( V3_NET_ERROR_NO_NET );
  }

  if ( ( net->status != V3_NET_STATUS_OPEN ) &&
    ( net->status != V3_NET_STATUS_LISTEN ) )
  {
    v3MutexUnlock( &net->lock );
    return( V3_PASS );
  }

#if ( defined( CPU_64BIT ) )
  socket_descriptor = net->handle;
#else
  socket_descriptor = v3u32u64( net->handle );
#endif
  if ( close( socket_descriptor ) < 0 )
  {
    /* Unable to close the socket */
    net->status = V3_NET_STATUS_CLOSED;
    net->last_error = V3_NET_ERROR_CLOSED;
    v3MutexUnlock( &net->lock );
    return( V3_NET_ERROR_CLOSED );
  }

  net->status = V3_NET_STATUS_CLOSED;
  net->last_error = V3_NET_ERROR_NONE;
  v3MutexUnlock( &net->lock );
  v3MutexFree( &net->lock );
  return( V3_PASS );
#endif /* NO_NETWORKING */
}

u32 v3NetDNS( u32 * ip, u32 length, ascii * name )
{
#if ( defined( NO_NETWORKING ) )
  return( 0 );
#else
  struct hostent *host_pointer;
  u32 count;

  if ( ( ip == NULL ) || ( length == 0 ) || ( name == NULL ) )
  {
    return( 0 );
  }

  if ( v3_NetGlobalInit() != V3_PASS )
  {
    return( 0 );
  }

  /* !!!: No thread lock! -- cramer */
  if ( ( host_pointer = gethostbyname( (char *) name ) ) == NULL )
  {
    /* Can't get IP of name */
    return( 0 );
  }

  /* read the returned list until we run into a NULL or hit length limit */
  for ( count = 0 ;
    ( ( length-- > 0 ) && ( host_pointer->h_addr_list[count] != NULL ) ) ;
    count++ )
  {
    /* net addresses come back in network byte order */
    v3u32Load( &ip[count], (const u8 *) host_pointer->h_addr_list[count] );
  }

  return( count );
#endif /* NO_NETWORKING */
}

s32 v3NetRevDNS( v3_NET_HOSTNAME * name, u32 ip )
{
#if ( defined( NO_NETWORKING ) )
  return( V3_FAIL );
#else
  struct hostent *host_pointer;
  u32 net_order;

  if ( ( name == NULL ) || ( ip == 0 ) )
  {
    return( V3_FAIL );
  }

  /* !!!: No thread lock! -- cramer */
  if ( v3_NetGlobalInit() != V3_PASS )
  {
    return( V3_FAIL );
  }

  v3u32Save( (u8 *) &net_order, &ip );

  if ( ( host_pointer = gethostbyaddr( (char *) &net_order,
    sizeof( net_order ), AF_INET ) ) == NULL )
  {
    /* Can't get a Domain Name for the IP */
    return( V3_FAIL );
  }

  /* Store the official domain name for the IP */
  v3StrCopyA( (ascii *) name, (ascii *) host_pointer->h_name,
    v3u64u32( V3_NET_MAX_HOSTNAME ) );

  return( V3_PASS );
#endif /* NO_NETWORKING */
}

s32 v3NetMyIP( u32 * ip )
{
#if ( defined( NO_NETWORKING ) )
  return( V3_FAIL );
#else
  v3_NET_HOSTNAME my_name;

  if ( ip == NULL )
  {
    return( V3_FAIL );
  }

  /* !!!: No thread lock! -- cramer */
  if ( v3_NetGlobalInit() != V3_PASS )
  {
    return( V3_FAIL );
  }

  /*
    This way is not the best with multi homed systems
    but we can return only one IP anyway
  */
  if ( gethostname( (char *) my_name, sizeof( my_name ) ) < 0 )
  {
    /* Can't get local hostname */
    return( V3_FAIL );
  }

  if ( ( v3NetDNS( ip, 1, my_name ) ) == 0 )
  {
    /* Can't get IP matching local hostname */
    return( V3_FAIL );
  }

  return( V3_PASS );
#endif /* NO_NETWORKING */
}

s32 v3NetACLAdd( v3_NET_ACL * acl, u32 ip, u32 mask, u32 permission,
  v3_time expires )
{
#if ( defined( NO_NETWORKING ) )
  return( V3_PASS );
#else
  u32 i;
  u64 size;
  void * mem;

  if ( ( acl == NULL ) || ( ip == 0 ) || ( ( permission != V3_NET_ALLOW ) &&
    ( permission != V3_NET_DENY ) ) )
  {
    return( V3_FAIL );
  }

  if ( ( acl->count + 1 ) > acl->length )
  {
    if ( acl->length == 0 )
    {
      /* 16 at a time */
      size = v3u64Lsh( v3u64u32( sizeof( v3_NET_ACL_ENTRY ) ), 4 );
      if ( ( mem = v3MemAlloc( size, V3_MEM_NORMAL ) ) == NULL )
      {
        return( V3_FAIL );
      }
      acl->list = mem;
      acl->length = 16;
    }
    else
    {
      /* 16 at a time */
      size = v3u64Mul( v3u64u32( sizeof( v3_NET_ACL_ENTRY ) ),
        v3u64u32( acl->length + 16 ) );
      if ( ( mem = v3MemRealloc( acl->list, size, V3_MEM_NORMAL ) ) == NULL )
      {
        return( V3_FAIL );
      }
      acl->list = mem;
      acl->length += 16;
    }
  }

  i = acl->count++;
  acl->list[i].expires = expires;
  acl->list[i].ip = ip;
  acl->list[i].mask = mask;
  acl->list[i].permission = permission;

  return( V3_PASS );
#endif
}

s32 v3NetACLDelete( v3_NET_ACL * acl, u32 ip, u32 mask )
{
#if ( defined( NO_NETWORKING ) )
  return( V3_PASS );
#else
  u32 i, j;
  v3_NET_ACL_ENTRY * list;

  if ( ( acl == NULL ) || ( ip == 0 ) )
  {
    return( V3_FAIL );
  }

  if ( acl->count == 0 )
  {
    return( V3_PASS );
  }

  list = acl->list;

  i = 0;
  do
  {
    if ( ( ip == list[i].ip ) && ( mask == list[i].mask ) )
    {
      for ( j = i ; j < ( acl->count - 1 ) ; j++ )
      {
        list[j] = list[j+1];
      }
      acl->count--;
    }
    else
    {
      i++;
    }
  } while ( i < acl->count );

  return( V3_PASS );
#endif
}

s32 v3NetACLTest( v3_NET_ACL * acl, u32 ip )
{
#if ( defined( NO_NETWORKING ) )
  return( V3_NET_DENY );
#else
  u32 i, j;
  s32 result;
  v3_NET_ACL_ENTRY * list;
  v3_time time;

  if ( acl == NULL )
  {
    return( V3_NET_ALLOW );
  }

  if ( acl->count == 0 )
  {
    return( V3_NET_ALLOW );
  }

  list = acl->list;

  if ( v3SystemClock( &time ) == V3_PASS )
  {
    i = 0;
    do
    {
      if ( v3s128Gt( list[i].expires, time ) )
      {
        for ( j = i ; j < ( acl->count - 1 ) ; j++ )
        {
          list[j] = list[j+1];
        }
        acl->count--;
      }
      else
      {
        i++;
      }
    } while ( i < acl->count );
  }

  result = V3_NET_ALLOW;
  for ( i = 0 ; i < acl->count ; i++ )
  {
    /* check for expires */
    if ( ( ip & list[i].mask ) == ( list[i].ip & list[i].mask ) )
    {
      result = list[i].permission;
    }
  }

  return( result );
#endif
}

void v3NetACLFree( v3_NET_ACL * acl )
{
#if ( !defined( NO_NETWORKING ) )

  if ( acl == NULL )
  {
    return;
  }

  acl->length = 0;
  acl->count = 0;

  v3MemFree( acl->list );
#endif
}

s32 v3_NetGlobalInit( void )
{
#if ( OS_TYPE == OS_WIN32 )
  static u32 net_global_init = 0;
  static WSADATA net_wsadata;

  if ( ( net_global_init == 0 ) &&
    ( WSAStartup( MAKEWORD( 1, 1 ), &net_wsadata ) != 0 ) )
  {
    return( V3_FAIL );
  }
  else
  {
    net_global_init = 1;
  }
#elif ( ( OS_TYPE == OS_LINUX ) || ( OS_TYPE == OS_TRU64 ) \
  || ( OS_TYPE == OS_FREEBSD ) || ( OS_TYPE == OS_SOLARIS ) \
  || ( OS_TYPE == OS_SUNOS ) )
  static u32 net_global_init = 0;

  if ( net_global_init == 0 )
  {
    signal( SIGPIPE, SIG_IGN );
    net_global_init = 1;
  }
#endif

  return( V3_PASS );
}

/* Testing */

s32 v3_TestCosmNet( void )
{
#if ( defined( NO_NETWORKING ) )
  return( V3_PASS );
#else
  u32 addr1, addr2;
  u32 i;
  v3_NET netsrv, netsrv1, netsrv2, netclient1, netclient2;
  v3_NET_ACL net_acl;
  ascii buf1[128], buf2[128];
  v3_NET_HOSTNAME host_name;

  if ( v3NetDNS( &addr2, 1, (ascii *) "localhost" ) == 0 )
  {
    return( -1 );
  }

  if ( addr2 != 0x7F000001 )
  { /* Should this be an error? */
    return( -2 );
  }

  v3MemSet( &netsrv, v3u64u32( sizeof( v3_NET ) ), 0 );
  v3MemSet( &netsrv1, v3u64u32( sizeof( v3_NET ) ), 0 );
  v3MemSet( &netsrv2, v3u64u32( sizeof( v3_NET ) ), 0 );
  v3MemSet( &netclient1, v3u64u32( sizeof( v3_NET ) ), 0 );
  v3MemSet( &netclient2, v3u64u32( sizeof( v3_NET ) ), 0 );
  v3MemSet( &net_acl, v3u64u32( sizeof( v3_NET_ACL ) ), 0 );
  v3MemSet( buf1, v3u64u32( sizeof( buf1 ) ), 0 );
  v3MemSet( buf2, v3u64u32( sizeof( buf2 ) ), 0 );

  for ( i = 0 ; i < 10 ; i++ )
  {
    buf1[i] = (ascii) i;
  }

  /*  Let the machine choose a free port.. */
  if ( v3NetListen( &netsrv, 0, 0, V3_NET_MODE_TCP, 1  ) != V3_PASS )
  {
    return( -3 );
  }

  /* Make sure we can't bind an already binded port.. */
  if ( v3NetListen( &netsrv1, 0, netsrv.my_port, V3_NET_MODE_TCP, 1 )
    == V3_PASS )
  {
    return( -4 );
  }

  /* Lets bind to a specific IP.. */
  if ( v3NetListen( &netsrv1, addr2, 0, V3_NET_MODE_TCP, 1 ) != V3_PASS )
  {
    return( -6 );
  }

  /* Unless MyIP is localhost */
#ifdef __FIX_THIS
  if ( addr1 != addr2 )
  {
    /* Connect to our server socket using the wrong IP. */
    if ( v3NetOpen( &netclient1, addr1, netsrv1->my_port, 0, V3_NET_MODE_TCP,
      0, 0, NULL ) == V3_PASS )
    {
      return( -7 );
    }

    /* Lets bind to the other IP, same port.. */
    if ( v3NetListen( &netsrv2, addr1, netsrv1->my_port, V3_NET_MODE_TCP, 1 )
      != V3_PASS )
    {
      return( -8 );
    }

    /* Lets try to close it */
    if ( v3NetClose( &netsrv2 ) != V3_PASS )
    {
      return( -9 );
    }
  }
#endif

  /* Lets try to close netsrv1 as well */
  if ( v3NetClose( &netsrv1 ) != V3_PASS )
  {
    return( -10 );
  }

  /* Connect to our server socket on localhost. */
  if ( v3NetOpen( &netclient1, addr2, netsrv.my_port, 0, 0, V3_NET_MODE_TCP,
    0, 0, (ascii *) NULL ) != V3_PASS )
  {
    return( -11 );
  }

  /* The above connection should be pending. */
  if ( v3NetAccept( &netsrv1, &netsrv, NULL, V3_NET_ACCEPT_NOWAIT )
    == V3_FAIL )
  {
    return( -12 );
  }

  /* Lets test pushing a buffer through. */
  if ( v3NetSend( &netclient1, buf1, 10 ) != V3_PASS )
  {
    return( -13 );
  }

  if ( v3NetRecv( buf2, 15, &netsrv1, 500 ) != 10 )
  {
    return( -14 );
  }

  for ( i = 0 ; i < 10 ; i++ )
  {
    if ( buf1[i] != buf2[i] )
    {
      return( -15 );
    }
  }

  /* Push some data 'backwards'. Change contents in buffer first.*/
  for ( i = 0 ; i < 10; i++ )
  {
      buf1[i] = (ascii) ( 10 - i );
  }

  if ( v3NetSend( &netsrv1, buf1, 10 ) != V3_PASS )
  {
    return( -16 );
  }

  /* and same again.. */
  if ( v3NetSend( &netsrv1, buf1, 10 ) != V3_PASS )
  {
    return( -17 );
  }

  if ( v3NetRecv( buf2, 25, &netclient1, 500 ) != 20 )
  {
    return( -18 );
  }

  for ( i = 0; i < 10; i++ )
  {
    if ( ( buf2[i] != ( 10 - i ) ) || ( buf2[i + 10] != ( 10 - i ) )
      || ( buf2[i + 20] == ( 10 - i ) ) )
    {
      return( -19 );
    }
  }

  /*
    Lets try reading where there should be no data. buf3 should be
    zero initialized.
  */
  if ( v3NetRecv( buf2, 100, &netsrv1, 500 ) > 0 )
  {
    /* Magically appearing data. No good. */
    return( -20 );
  }

  /* netsrv2's mode never gets set. Set it to TCP. */
  netsrv2.mode = V3_NET_MODE_TCP;

  /* Nothing should be pending. */
  if ( v3NetAccept( &netsrv2, &netsrv, NULL, V3_NET_ACCEPT_NOWAIT )
    == V3_PASS )
  {
    return( -21 );
  }

  /* Open up a second 'client' against the 'server'. */
  if ( _V3_NETOPEN( &netclient2, addr2, netsrv.my_port ) != V3_PASS )
  {
    return( -22 );
  }

  /* Accept the second connection. */
  if ( v3NetAccept( &netsrv2, &netsrv, NULL, V3_NET_ACCEPT_NOWAIT )
    == V3_FAIL )
  {
    return( -23 );
  }

  /*
    Try sending some stuff on the new line. Make sure it doesn't
    echo to any other lines.
  */
  for ( i = 0 ; i < 15 ; i++ )
  {
    buf1[i] = (ascii) ( i * 2 );
  }

  if ( v3NetSend( &netclient2, buf1, 15 ) != V3_PASS )
  {
    return( -24 );
  }

  /* Check the three sockets there should -not- be data on. */
  if ( v3NetRecv( buf2, 100, &netsrv1, 500 ) > 0 )
  {
    return( -25 );
  }

  if ( v3NetRecv( buf2, 100, &netclient1, 500 ) > 0 )
  {
    return( -26 );
  }

  if ( v3NetRecv( buf2, 100, &netclient2, 500 ) > 0 )
  {
    return( -27 );
  }

  /* Now take in proper data. */
  if ( v3NetRecv( buf2, 100, &netsrv2, 500 ) != 15 )
  {
    return( -28 );
  }

  for ( i = 0 ; i < 15 ; i++ )
  {
    if ( ( buf2[i] != buf1[i] ) || ( buf2[i] != (ascii) ( i * 2 ) ) )
    {
      return( -29 );
    }
  }

  /* Close, cleanup. */
  if ( v3NetClose( &netsrv1 ) != V3_PASS )
  {
    return( -30 );
  }

  if ( v3NetClose( &netsrv2 ) != V3_PASS )
  {
    return( -31 );
  }

  if ( v3NetClose( &netclient1 ) != V3_PASS )
  {
    return( -32 );
  }

  /* this one is tricky, it should fail on the second call */
  if ( ( v3NetSend( &netclient2, buf1, 10 ) == V3_PASS )
    && ( v3NetSend( &netclient2, buf1, 10 ) == V3_PASS ) )
  {
    return( -33 );
    /* most likely v3_NetGlobalInit failure */
  }

/*
  if ( v3NetClose( &netclient2 ) != V3_PASS )
  {
    return( -34 );
  }
*/

  if ( v3NetClose( &netsrv ) != V3_PASS )
  {
    return( -35 );
  }

#ifdef __FIX_THIS
  if ( v3NetRevDNS( (v3_NET_HOSTNAME *) host_name, addr2 ) == V3_FAIL )
  {
    return( -36 );
  }

  if ( strncmp( (char *) host_name, "localhost", 9 ) != 0 )
  {
    return( -37 );
  }

  if ( v3NetMyIP( &addr1 ) == V3_FAIL )
  {
    /* we claim to have a network, so we should have one */
    return( -38 );
  }
#endif

  /*
    Net ACL tests
  */

  return( V3_PASS );
#endif /* NO_NETWORKING */
}
