/*
  Use of this code is subject to the terms of the Phase 1 Cosm License.
  Any other use of this code violates the terms of the license.
*/

/*
  We put all error I/O here because it will significantly increase
  the size of the code. This code will not end up in libCosm.
  Ironicly, the only function we can't test (v3PrintA) is the only
  one we use here.
*/

#include "cosm.h"

typedef struct
{
  u64 a;
  u32 b;
  u16 c;
  u8  d[2];
  u128 e;
  u8  f[2];
  u16 g;
  u32 h;
  u64 i;
} align_test; /* 48 bytes */

void CatchINT( int arg )
{
  v3SignalRegister( V3_PROCESS_SIGINT, CatchINT );
  v3PrintA( (ascii *) "\nSIGINT - ouch!\n" );
}

void CatchTERM( int arg )
{
  v3SignalRegister( V3_PROCESS_SIGTERM, CatchTERM );
  v3PrintA( (ascii *) "\nSIGTERM - ouch!\n" );
}

void ChildINT( int arg )
{
  v3SignalRegister( V3_PROCESS_SIGINT, ChildINT );
  v3PrintA( (ascii *) "\nChild SIGINT - ouch!\n" );
}

void ChildTERM( int arg )
{
  v3SignalRegister( V3_PROCESS_SIGTERM, ChildTERM );
  v3PrintA( (ascii *) "\nChild SIGTERM - ouch!\n" );
}

void Child( void )
{
  v3PrintA( (ascii *) "Hello from child process %v\n", v3ProcessID() );

  if ( ( v3SignalRegister( V3_PROCESS_SIGINT, ChildINT ) != V3_PASS ) ||
    ( v3SignalRegister( V3_PROCESS_SIGTERM, ChildTERM ) != V3_PASS ) )
  {
    v3PrintA( (ascii *) "Child Signal registration failed\n" );
    v3ProcessEnd( -1 );
  }

  /* sleep long enough for tests */
  v3Sleep( 10000 );
  /* !!! UNIX HACK, need to fix v3Sleep/usleep */
  v3Sleep( 10000 );
  v3Sleep( 10000 );
  v3Sleep( 10000 );


  v3ProcessEnd( 0 );
}

int main( int argc, char * argv[] )
{
  s32 error;
  s32 error2;
  const ascii * os_types[] = V3_OS_TYPES;
  const ascii * cpu_types[] = V3_CPU_TYPES;
  const ascii * test_modules[] = V3_TEST_MODULES;
  const ascii * months[12] = V3_TIME_MONTHS;
  const ascii * days[7] = V3_TIME_DAYS;
  u32 cpuwords;
  u64 memory;
  u64 length;
  ascii buffer[256];
  v3_time mytime;
  v3_TIME_UNITS myunits;
  u64 milli;
  u32 count;
#if ( !defined( NO_IEEE_FLOAT ) )
  f64 fp;
#endif
  v3_LOG log;
  v3_FILE file;
  v3_FILE_INFO info;
  ascii * correct, buf[512];
  u64 bytesread;
  u64 process_id;
  u64 child_id;

  if ( ( argc == 2 ) && ( v3StrCmpA( (ascii * ) argv[1],
    (ascii *) "-child", v3u64u32( 8 ) ) == 0 ) )
  {
    Child();
  }

  count = 0;
  while ( os_types[count] != NULL )
  {
    count++;
  }
  if ( count != ( V3_OS_TYPE_MAX + 1 ) )
  {
    v3PrintA( (ascii *) "V3_OS_TYPES are incorrect.\n" );
  }

  count = 0;
  while ( cpu_types[count] != NULL )
  {
    count++;
  }
  if ( count != ( V3_CPU_TYPE_MAX + 1 ) )
  {
    v3PrintA( (ascii *) "V3_CPU_TYPES are incorrect.\n" );
  }

#if ( defined( CPU_64BIT ) )
  v3PrintA( (ascii *) "64-Bit CPU, " );
#else
  v3PrintA( (ascii *) "32-Bit CPU, " );
#endif

  v3PrintA( (ascii *) "OS = %.20s, CPU = %.20s * %u, ",
    os_types[OS_TYPE], cpu_types[CPU_TYPE], v3CPUCount() );

  cpuwords = ( ( (u32) sizeof( long ) * 100 ) +
    ( (u32) sizeof( int ) * 10 ) + (u32) sizeof( short ) );
  v3PrintA( (ascii *) "CPU_WORD_SIZES set/detected : %u/%u\n",
    CPU_WORD_SIZES, cpuwords );
  if ( CPU_WORD_SIZES != cpuwords )
  {
    v3PrintA( (ascii *)
      "CPU_WORD_SIZES is set incorrectly, fix the make/os-cpu makefile.\n" );
  }
  if ( sizeof( align_test ) != 48 )
  {
    v3PrintA( (ascii *) "Data alignment is wrong, check the compiler.\n" );
  }

  v3PrintA( (ascii *) "Running system tests... " );
  if ( v3Test( &error, &error2, 0 ) != V3_PASS )
  {
    v3PrintA( (ascii *) "Test failure in module %.16s %i.\n",
      test_modules[-error], error2 );
    v3ProcessEnd( error );
  }
  v3PrintA( (ascii *) "all passed.\n" );

  if ( v3MemSystem( &memory ) == V3_PASS )
  {
    v3PrintA( (ascii *) "Memory detected: %v bytes (~%v MB)\n", memory,
      v3u64Rsh( memory, 20 ) );
  }

#if ( !defined( NO_IEEE_FLOAT ) )
  fp = (f64) 3.1415926535897932384626433832795;
  v3PrintA( (ascii *)
    "IEEE floating point enabled\n  PI = %.8g ~= [3.14159265]\n", fp );
  fp = sin( (f64) 1.5707963267948966192313216916398 );
  v3PrintA( (ascii *) "  sin( PI/2 ) = %.6g ~= [1.000000]\n\n", fp );
#endif

  v3PrintA( (ascii *) "Run ANSI color test? [y/N] " );
  _V3_SET64( length, 00000000, 00000004 );
  v3InputA( buffer, length, V3_IO_ECHO );
  if ( ( buffer[0] == 'y' ) || ( buffer[0] == 'Y' ) )
  {
    v3PrintA( (ascii *) "%.5s%.5sS%.5s%.5sP%.5s%.5sE%.5s%.5sC",
      ANSI_BG_WHITE, ANSI_FG_BLACK, ANSI_BG_MAGENTA, ANSI_FG_RED,
      ANSI_BG_BLUE, ANSI_FG_YELLOW, ANSI_BG_CYAN, ANSI_FG_GREEN );
    v3PrintA( (ascii *) "%.5s%.5sT%.5s%.5sR%.5s%.5sU%.5s%.5sM",
      ANSI_BG_GREEN, ANSI_FG_CYAN, ANSI_BG_YELLOW, ANSI_FG_BLUE,
      ANSI_BG_RED, ANSI_FG_MAGENTA, ANSI_BG_BLACK, ANSI_FG_WHITE );
    v3PrintA( (ascii *) "%.5s\n\n", ANSI_CLEAR );
  }

  _V3_SET64( length, 00000000, 00000010 );
  v3PrintA( (ascii *)
    "Type 20 characters, 15 will be read (should be echoed): " );
  v3InputA( buffer, length, V3_IO_ECHO );
  v3PrintA( (ascii *) "You typed \"%.255s\"\n", buffer );
  v3PrintA( (ascii *)
    "Type 20 characters, 15 will be read (should not be echoed): " );
  v3InputA( buffer, length, V3_IO_NOECHO );
  v3PrintA( (ascii *) "You typed \"%.255s\"\n", buffer );
  v3PrintA( (ascii *)
    "\nThe following 6 times should be about 1/2 second apart:\n" );
  _V3_SET64( milli, 00418937, 4BC6A7EF )
  for ( count = 0 ; count < 6 ; count++ )
  {
    if ( v3SystemClock( &mytime ) == V3_PASS )
    {
      if ( v3TimeUnitsGregorian( &myunits, mytime ) == V3_PASS )
      {
        v3PrintA( (ascii *) "%.9s %.9s %u, %j %02u:%02u:%02u.%03v UTC, d%u\n",
          days[myunits.wday], months[myunits.month], myunits.day + 1,
          myunits.year, myunits.hour, myunits.min, myunits.sec,
          v3u64Div( myunits.subsec, milli ), myunits.yday + 1 );
        v3Sleep( 500 );
      }
    }
  }

  /* test the log echo's */
  v3MemSet( &log, v3u64u32( sizeof( v3_LOG ) ), 0 );
  v3MemSet( &file, v3u64u32( sizeof( v3_FILE ) ), 0 );
  if ( v3LogOpen( &log, (ascii *) "testlib.log", 5, V3_LOG_MODE_NUMBER )
    != V3_PASS )
  {
    v3PrintA( (ascii *) "Unable to open test log.\n");
  }
  else
  {
    v3PrintA( (ascii *) "\nThis should be logged and echoed: " );
    v3Log( &log, 3, V3_LOG_ECHO, (ascii *) "Logged and echoed.\n" );
    v3PrintA( (ascii *) "This should be logged and NOT echoed: " );
    v3Log( &log, 3, V3_LOG_NOECHO, (ascii *) "Logged and not echoed.\n" );
    v3PrintA( (ascii *) "\n");
    v3LogClose( &log );

    correct = (ascii *) "Logged and echoed.\nLogged and not echoed.\n";
    if ( v3FileOpen( &file, (ascii *) "testlib.log", V3_FILE_MODE_READ,
      V3_FILE_LOCK_NONE ) != V3_PASS )
    {
      v3PrintA( (ascii *) "Unable to open test log for checking.\n");
    }
    else
    {
      bytesread = v3FileRead( buf, v3StrLengthA( correct ), &file );

      if ( v3StrCmpA( buf, correct, v3StrLengthA( correct ) ) != V3_PASS )
      {
        v3PrintA( (ascii *) "Error in logging/echo.\n");
      }
      else
      {
        v3PrintA( (ascii *) "Logging/echo correct.\n");
      }
      v3FileClose( &file );
    }
  }

  v3FileDelete( (ascii *) "testlib.log" );

  /* test fileinfo */
  v3FileInfo( &info, (ascii *) argv[0] );
  v3PrintA( (ascii *)
    "\ntestlib.exe info:\nLength: %v\nType: %u\nRights: %u\n",
    info.length, info.type, info.rights );
  if ( v3TimeUnitsGregorian( &myunits, info.create ) == V3_PASS )
  {
    v3PrintA( (ascii *)
      "Create: %.9s %.9s %u, %j %02u:%02u:%02u.%03v UTC, d%u\n",
      days[myunits.wday], months[myunits.month], myunits.day + 1,
      myunits.year, myunits.hour, myunits.min, myunits.sec,
      v3u64Div( myunits.subsec, milli ), myunits.yday + 1 );
  }
  if ( v3TimeUnitsGregorian( &myunits, info.modify ) == V3_PASS )
  {
    v3PrintA( (ascii *)
      "Modify: %.9s %.9s %u, %j %02u:%02u:%02u.%03v UTC, d%u\n",
      days[myunits.wday], months[myunits.month], myunits.day + 1,
      myunits.year, myunits.hour, myunits.min, myunits.sec,
      v3u64Div( myunits.subsec, milli ), myunits.yday + 1 );
  }
  if ( v3TimeUnitsGregorian( &myunits, info.access ) == V3_PASS )
  {
    v3PrintA( (ascii *)
      "Access: %.9s %.9s %u, %j %02u:%02u:%02u.%03v UTC, d%u\n",
      days[myunits.wday], months[myunits.month], myunits.day + 1,
      myunits.year, myunits.hour, myunits.min, myunits.sec,
      v3u64Div( myunits.subsec, milli ), myunits.yday + 1 );
  }

  v3PrintA( (ascii *) "\nRun process spawning test? [y/N] " );
  _V3_SET64( length, 00000000, 00000010 );
  v3InputA( buffer, length, V3_IO_ECHO );
  if ( ( buffer[0] == 'y' ) || ( buffer[0] == 'Y' ) )
  {
    v3PrintA( (ascii *) "Spawning - %.256s -child\n", argv[0] );
    child_id = v3ProcessSpawn( (ascii *) argv[0], (ascii *) "-child",
      NULL );
    if ( _V3_EQ64( child_id, 00000000, 00000000 ) )
    {
      v3PrintA( (ascii *) "Spawn failed\n" );
    }
    else
    {
      v3PrintA( (ascii *) "Spawned PID %v\n", child_id );
    }

    /* allow process time to start */
    v3Sleep( 500 );
    if ( v3ProcessSignal( child_id, V3_PROCESS_SIGPING ) != V3_PASS )
    {
      v3PrintA( (ascii *) "\nChild SIGPING failed (not started?)\n");
    }
    else
    {
      v3PrintA( (ascii *) "\nChild SIGPING worked (exists)\n");
    }

    v3PrintA( (ascii *) "Sending Child SIGINT\n");
    if ( v3ProcessSignal( child_id, V3_PROCESS_SIGINT ) != V3_PASS )
    {
      v3PrintA( (ascii *) "Child SIGINT failed\n");
    }
    else
    {
      v3PrintA( (ascii *) "Child SIGINT worked\n");
    }

    v3Sleep( 500 );
    v3PrintA( (ascii *) "Sending Child SIGTERM\n");
    if ( v3ProcessSignal( child_id, V3_PROCESS_SIGTERM ) != V3_PASS )
    {
      v3PrintA( (ascii *) "Child SIGTERM failed\n");
    }
    else
    {
      v3PrintA( (ascii *) "Child SIGTERM worked\n");
    }

    v3Sleep( 500 );
    v3PrintA( (ascii *) "Sending Child SIGKILL (should kill it)\n");
    if ( v3ProcessSignal( child_id, V3_PROCESS_SIGKILL ) != V3_PASS )
    {
      v3PrintA( (ascii *) "Child SIGKILL failed\n");
    }
    else
    {
      v3PrintA( (ascii *) "Child SIGKILL worked\n");
    }

    v3Sleep( 500 );
    if ( v3ProcessSignal( child_id, V3_PROCESS_SIGPING ) == V3_PASS )
    {
      v3PrintA( (ascii *) "\nChild SIGPING failed, thinks child exists\n" );
    }
    else
    {
      v3PrintA( (ascii *) "\nChild SIGPING worked (it's dead Jim)\n");
    }
  }

  if ( ( v3SignalRegister( V3_PROCESS_SIGINT, CatchINT ) != V3_PASS ) ||
    ( v3SignalRegister( V3_PROCESS_SIGTERM, CatchTERM ) != V3_PASS ) )
  {
    v3PrintA( (ascii *) "Signal registration failed\n" );
    v3ProcessEnd( -1 );
  }

  process_id = v3ProcessID();
  if ( v3ProcessSignal( process_id, V3_PROCESS_SIGPING ) != V3_PASS )
  {
    v3PrintA( (ascii *) "\nSIGPING failed\n");
    v3ProcessEnd( -1 );
  }
  else
  {
    v3PrintA( (ascii *) "\nSIGPING worked\n");
  }

  v3Sleep( 500 );
  v3PrintA( (ascii *) "Sending SIGINT\n");
  if ( v3ProcessSignal( process_id, V3_PROCESS_SIGINT ) != V3_PASS )
  {
    v3PrintA( (ascii *) "SIGINT failed\n");
  }
  else
  {
    v3PrintA( (ascii *) "SIGINT worked\n");
  }

  v3Sleep( 500 );
  v3PrintA( (ascii *) "Sending SIGTERM\n");
  if ( v3ProcessSignal( process_id, V3_PROCESS_SIGTERM ) != V3_PASS )
  {
    v3PrintA( (ascii *) "SIGTERM failed\n");
  }
  else
  {
    v3PrintA( (ascii *) "SIGTERM worked\n");
  }

  v3Sleep( 500 );
  v3PrintA( (ascii *) "Sending SIGKILL (program should end)\n");
  if ( v3ProcessSignal( process_id, V3_PROCESS_SIGKILL ) != V3_PASS )
  {
    v3PrintA( (ascii *) "SIGKILL failed\n");
  }

  /* if SIGKILL worked, we're dead by now */
  v3Sleep( 20 );
  v3PrintA( (ascii *) "SIGKILL failed, still alive\n");

  return( 0 );
}
